#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "mpi.h"

int main( int argc, char *argv[] )
{   
    int myrank, nprocs;
    int n, mat_size;
    int i;
    float *sm1, *sm2; // SM: submatrix
    MPI_Request request;

    MPI_Init(&argc,&argv);

    if ( argc != 2 )
    {
        fprintf( stderr, "Usage: %s n\n", argv[0] );
        exit( -1 );
    }

    n = atoi(argv[1]);
    mat_size = (n/2) * (n/2);

    MPI_Comm_rank( MPI_COMM_WORLD, &myrank );
    MPI_Comm_size( MPI_COMM_WORLD, &nprocs );

    sm1 = (float *) malloc ( mat_size * sizeof(float) );
    sm2 = (float *) malloc ( mat_size * sizeof(float) );

    // Initialize my submatrix of A (sm1)
    for ( i = 0; i < mat_size; i++ )
        sm1[ i ] = myrank;

    // Communication step 1 ( row-wise )
    if ( myrank % 2 == 0 ) // I'm column 0
    {
        MPI_Isend( sm1, mat_size, MPI_FLOAT, myrank+1, 0, MPI_COMM_WORLD, &request ); // Send my chunk of A
        MPI_Recv( sm2, mat_size, MPI_FLOAT, myrank+1, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE ); // Recv in sm2
        MPI_Wait( &request, MPI_STATUS_IGNORE );
    }
    else // I'm column 1
    {
        MPI_Isend( sm1, mat_size, MPI_FLOAT, myrank-1, 0, MPI_COMM_WORLD, &request ); // Send my chunk of A
        MPI_Recv( sm2, mat_size, MPI_FLOAT, myrank-1, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE ); // Recv in sm2
        MPI_Wait( &request, MPI_STATUS_IGNORE );
    }

    // Operate
    if ( myrank % 2 == 0 ) // I'm column 0, addition
        for ( i = 0; i < mat_size; i++ )
            sm1[i] = sm1[i] + sm2[i]; // sm1: B
    else // I'm column 1, substraction
        for ( i = 0; i < mat_size; i++ )
            sm1[i] = sm2[i] - sm1[i]; // sm1: B

    // Communication step 2
    if ( myrank / 2 == 0 ) // I'm row 0
    {
        MPI_Isend( sm1, mat_size, MPI_FLOAT, myrank+2, 1, MPI_COMM_WORLD, &request ); // Send my chunk of B
        MPI_Recv( sm2, mat_size, MPI_FLOAT, myrank+2, 1, MPI_COMM_WORLD, MPI_STATUS_IGNORE ); // Recv in sm2
        MPI_Wait( &request, MPI_STATUS_IGNORE );
    }
    else // I'm row 1
    {
        MPI_Isend( sm1, mat_size, MPI_FLOAT, myrank-2, 1, MPI_COMM_WORLD, &request ); // Send my chunk of B
        MPI_Recv( sm2, mat_size, MPI_FLOAT, myrank-2, 1, MPI_COMM_WORLD, MPI_STATUS_IGNORE ); // Recv in sm2
        MPI_Wait( &request, MPI_STATUS_IGNORE );
    }

    // Operate
    if ( myrank / 2 == 0 ) // I'm row 0, addition
        for ( i = 0; i < mat_size; i++ )
            sm1[i] = sm1[i] + sm2[i]; // sm1: C
    else // I'm row 1, substraction
        for ( i = 0; i < mat_size; i++ )
            sm1[i] = sm2[i] - sm1[i]; // sm1: C

    // Check result
    float res[4] = { 6, -2, -4, 0};
    int myres = res[myrank];
    int error = 0;
    for ( i = 0; i < mat_size; i++ )
        sm1[i] -= myres;
    for ( i = 0; i < mat_size; i++ )
        if ( fabs( sm1[i] ) > 1e-5 ) // some threshold
        {
            fprintf( stderr, "[%d] %d: %.7f\n", myrank, i, sm1[i] );
            error = 1;
            break;
        }

    if ( error )
        printf("[%d] ERROR!\n", myrank);
    else
        printf("[%d] WORKED!\n", myrank);

    free( sm1 );
    free( sm2 );

    MPI_Finalize();

    return 0;
}
